@@ -24,7 +24,6 @@
 #  Based on libpodcasts.py (thp, 2005-10-29)
 #
 
-import collections
 import datetime
 import glob
 import hashlib
@@ -47,39 +46,39 @@ _ = gpodder.gettext
 
 
 class Feed:
-    """ abstract class for presenting a parsed feed to PodcastChannel """
+    """Abstract class for presenting a parsed feed to PodcastChannel."""
 
     def get_title(self):
-        """ :return str: the feed's title """
+        """Return the feeds title or None."""
         return None
 
     def get_link(self):
-        """ :return str: link to the feed's website """
+        """Return the link to the feeds website or None."""
         return None
 
     def get_description(self):
-        """ :return str: feed's textual description """
+        """Return the feeds textual description or None."""
         return None
 
     def get_cover_url(self):
-        """ :return str: url of the feed's cover image """
+        """Return the URL of the feeds cover image or None."""
         return None
 
     def get_payment_url(self):
-        """ :return str: optional -- feed's payment url """
+        """Return the feeds payment url or None."""
         return None
 
     def get_http_etag(self):
-        """ :return str: optional -- last HTTP etag header, for conditional request next time """
+        """Return the last HTTP etag header, for conditional request next time, or None."""
         return None
 
     def get_http_last_modified(self):
-        """ :return str: optional -- last HTTP Last-Modified header, for conditional request next time """
+        """Return the last HTTP Last-Modified header, for conditional request next time, or None."""
         return None
 
     def get_new_episodes(self, channel, existing_guids):
-        """
-        Produce new episodes and update old ones.
+        """Produce new episodes and update old ones.
+
         Feed is a class to present results, so the feed shall have already been fetched.
         Existing episodes not in all_seen_guids will be purged from the database.
         :param PodcastChannel channel: the updated channel
@@ -89,8 +88,8 @@ class Feed:
         return ([], set())
 
     def get_next_page(self, channel, max_episodes):
-        """
-        Paginated feed support (RFC 5005).
+        """Paginated feed support (RFC 5005).
+
         If the feed is paged, return the next feed page.
         Returned page will in turn be asked for the next page, until None is returned.
         :return feedcore.Result: the next feed's page,
@@ -198,10 +197,8 @@ class PodcastParserFeed(Feed):
 
 
 class gPodderFetcher(feedcore.Fetcher):
-    """
-    This class implements fetching a channel from custom feed handlers
-    or the default using podcastparser
-    """
+    """Implements fetching a channel from custom feed handlers or the default using podcastparser."""
+
     def fetch_channel(self, channel, max_episodes):
         custom_feed = registry.feed_handler.resolve(channel, None, max_episodes)
         if custom_feed is not None:
@@ -242,17 +239,16 @@ class gPodderFetcher(feedcore.Fetcher):
 
 
 class PodcastModelObject(object):
-    """
-    A generic base class for our podcast model providing common helper
-    and utility functions.
-    """
+    """A generic base class for our podcast model providing common helper and utility functions."""
+
     __slots__ = ('id', 'parent', 'children')
 
     @classmethod
     def create_from_dict(cls, d, *args):
-        """
-        Create a new object, passing "args" to the constructor
-        and then updating the object with the values from "d".
+        """Create a podcast model from constructor args and dict.
+
+        Passes "args" to the constructor and then updates the object with
+        the values from "d".
         """
         o = cls(*args)
 
@@ -264,7 +260,8 @@ class PodcastModelObject(object):
 
 
 class PodcastEpisode(PodcastModelObject):
-    """holds data for one object in a channel"""
+    """Holds data for one object in a channel."""
+
     # In theory, Linux can have 255 bytes (not characters!) in a filename, but
     # filesystems like eCryptFS store metadata in the filename, making the
     # effective number of characters less than that. eCryptFS recommends
@@ -283,8 +280,8 @@ class PodcastEpisode(PodcastModelObject):
     is_locked = property(fget=_deprecated, fset=_deprecated)
 
     def has_website_link(self):
-        return bool(self.link) and (self.link != self.url or
-                youtube.is_video_link(self.link))
+        return bool(self.link) and (self.link != self.url
+                or youtube.is_video_link(self.link))
 
     @classmethod
     def from_podcastparser_entry(cls, entry, channel):
@@ -293,6 +290,11 @@ class PodcastEpisode(PodcastModelObject):
         episode.title = entry['title']
         episode.link = entry['link']
         episode.episode_art_url = entry.get('episode_art_url')
+
+        # Only one of the two description fields should be set at a time.
+        # This keeps the database from doubling in size and reduces load time from slow storage.
+        # episode._text_description is initialized by episode.cache_text_description() from the set field.
+        # episode.html_description() returns episode.description_html or generates from episode.description.
         if entry.get('description_html'):
             episode.description = ''
             episode.description_html = entry['description_html']
@@ -318,6 +320,7 @@ class PodcastEpisode(PodcastModelObject):
             episode.url = _url
         media_available = audio_available or video_available or link_has_media
 
+        url_is_invalid = False
         for enclosure in entry['enclosures']:
             episode.mime_type = enclosure['mime_type']
 
@@ -333,6 +336,7 @@ class PodcastEpisode(PodcastModelObject):
 
             episode.url = util.normalize_feed_url(enclosure['url'])
             if not episode.url:
+                url_is_invalid = True
                 continue
 
             episode.file_size = enclosure['file_size']
@@ -341,7 +345,13 @@ class PodcastEpisode(PodcastModelObject):
         # Brute-force detection of the episode link
         episode.url = util.normalize_feed_url(entry['link'])
         if not episode.url:
-            return None
+            # The episode has no downloadable content.
+            # Set an empty URL so downloading will fail.
+            episode.url = ''
+            # Display an error icon if URL is invalid.
+            if url_is_invalid or (entry['link'] is not None and entry['link'] != ''):
+                episode._download_error = 'Invalid episode URL'
+            return episode
 
         if any(mod.is_video_link(episode.url) for mod in (youtube, vimeo)):
             return episode
@@ -408,7 +418,7 @@ class PodcastEpisode(PodcastModelObject):
 
     @property
     def trimmed_title(self):
-        """Return the title with the common prefix trimmed"""
+        """Return the title with the common prefix trimmed."""
         # Minimum amount of leftover characters after trimming. This
         # avoids things like "Common prefix 123" to become just "123".
         # If there are LEFTOVER_MIN or less characters after trimming,
@@ -418,8 +428,8 @@ class PodcastEpisode(PodcastModelObject):
         # "Podcast Name - Title" and "Podcast Name: Title" -> "Title"
         for postfix in (' - ', ': '):
             prefix = self.parent.title + postfix
-            if (self.title.startswith(prefix) and
-                    len(self.title) - len(prefix) > LEFTOVER_MIN):
+            if (self.title.startswith(prefix)
+                    and len(self.title) - len(prefix) > LEFTOVER_MIN):
                 return self.title[len(prefix):]
 
         regex_patterns = [
@@ -438,14 +448,14 @@ class PodcastEpisode(PodcastModelObject):
 
         # "#001: Title" -> "001: Title"
         if (
-                not self.parent._common_prefix and
-                re.match(r'^#\d+: ', self.title) and
-                len(self.title) - 1 > LEFTOVER_MIN):
+                not self.parent._common_prefix
+                and re.match(r'^#\d+: ', self.title)
+                and len(self.title) - 1 > LEFTOVER_MIN):
             return self.title[1:]
 
-        if (self.parent._common_prefix is not None and
-                self.title.startswith(self.parent._common_prefix) and
-                len(self.title) - len(self.parent._common_prefix) > LEFTOVER_MIN):
+        if (self.parent._common_prefix is not None
+                and self.title.startswith(self.parent._common_prefix)
+                and len(self.title) - len(self.parent._common_prefix) > LEFTOVER_MIN):
             return self.title[len(self.parent._common_prefix):]
 
         return self.title
@@ -477,13 +487,20 @@ class PodcastEpisode(PodcastModelObject):
         return player
 
     def can_play(self, config):
-        """
-        # gPodder.playback_episodes() filters selection with this method.
-        """
-        return self.was_downloaded(and_exists=True) or self.can_stream(config)
+        """gPodder.playback_episodes() filters selection with this method."""
+        return (self.was_downloaded(and_exists=True)
+                or self.can_preview()
+                or self.can_stream(config))
+
+    def can_preview(self):
+        return (self.downloading
+                and self.download_task.custom_downloader is not None
+                and self.download_task.custom_downloader.partial_filename is not None
+                and os.path.exists(self.download_task.custom_downloader.partial_filename))
 
     def can_stream(self, config):
-        """
+        """Return True if episode can be streamed.
+
         Don't try streaming if the user has not defined a player
         or else we would probably open the browser when giving a URL to xdg-open.
         We look at the audio or video player depending on its file type.
@@ -492,44 +509,44 @@ class PodcastEpisode(PodcastModelObject):
         return player and player != 'default'
 
     def can_download(self):
-        """
+        """Return True if episode can be downloaded.
+
         gPodder.on_download_selected_episodes() filters selection with this method.
         PAUSING and PAUSED tasks can be resumed.
         """
         return not self.was_downloaded(and_exists=True) and (
-            not self.download_task
+            self.download_task is None
             or self.download_task.can_queue()
             or self.download_task.status == self.download_task.PAUSING)
 
     def can_pause(self):
-        """
-        gPodder.on_pause_selected_episodes() filters selection with this method.
-        """
-        return self.download_task and self.download_task.can_pause()
+        """gPodder.on_pause_selected_episodes() filters selection with this method."""
+        return self.download_task is not None and self.download_task.can_pause()
 
     def can_cancel(self):
-        """
-        DownloadTask.cancel() only cancels the following tasks.
-        """
-        return self.download_task and self.download_task.can_cancel()
+        """DownloadTask.cancel() only cancels the following tasks."""
+        return self.download_task is not None and self.download_task.can_cancel()
 
     def can_delete(self):
-        """
-        gPodder.delete_episode_list() filters out locked episodes, and cancels all unlocked tasks in selection.
+        """Return True, if episode can be deleted.
+
+        gPodder.delete_episode_list() filters out locked episodes,
+        and cancels all unlocked tasks in selection.
         """
         return self.state != gpodder.STATE_DELETED and not self.archive and (
-            not self.download_task or self.download_task.status == self.download_task.FAILED)
+            self.download_task is None or self.download_task.status == self.download_task.FAILED)
 
     def can_lock(self):
-        """
+        """Return True, if episode can be locked.
+
         gPodder.on_item_toggle_lock_activate() unlocks deleted episodes and toggles all others.
         Locked episodes can always be unlocked.
         """
         return self.state != gpodder.STATE_DELETED or self.archive
 
     def check_is_new(self):
-        return (self.state == gpodder.STATE_NORMAL and self.is_new and
-                not self.downloading)
+        return (self.state == gpodder.STATE_NORMAL and self.is_new
+                and not self.downloading)
 
     def save(self):
         gpodder.user_extensions.on_episode_save(self)
@@ -612,7 +629,7 @@ class PodcastEpisode(PodcastModelObject):
         self.set_state(gpodder.STATE_DELETED)
 
     def get_playback_url(self, config=None, allow_partial=False):
-        """Local (or remote) playback/streaming filename/URL
+        """Local (or remote) playback/streaming filename/URL.
 
         Returns either the local filename or a streaming URL that
         can be used to playback this episode.
@@ -620,11 +637,10 @@ class PodcastEpisode(PodcastModelObject):
         Also returns the filename of a partially downloaded file
         in case partial (preview) playback is desired.
         """
-        url = self.local_filename(create=False)
+        if (allow_partial and self.can_preview()):
+            return self.download_task.custom_downloader.partial_filename
 
-        if (allow_partial and url is not None and
-                os.path.exists(url + '.partial')):
-            return url + '.partial'
+        url = self.local_filename(create=False)
 
         if url is None or not os.path.exists(url):
             # FIXME: may custom downloaders provide the real url ?
@@ -636,13 +652,13 @@ class PodcastEpisode(PodcastModelObject):
         filename = filename.strip('.' + string.whitespace) + extension
 
         for name in util.generate_names(filename):
-            if (not self.db.episode_filename_exists(self.podcast_id, name) or
-                    self.download_filename == name):
+            if (not self.db.episode_filename_exists(self.podcast_id, name)
+                    or self.download_filename == name):
                 return name
 
     def local_filename(self, create, force_update=False, check_only=False,
             template=None, return_wanted_filename=False):
-        """Get (and possibly generate) the local saving filename
+        """Get (and possibly generate) the local saving filename.
 
         Pass create=True if you want this function to generate a
         new filename if none exists. You only want to do this when
@@ -700,9 +716,9 @@ class PodcastEpisode(PodcastModelObject):
                 episode_filename, _ = util.filename_from_url(url)
 
             # Use title for YouTube, Vimeo and Soundcloud downloads
-            if (youtube.is_video_link(self.url) or
-                    vimeo.is_video_link(self.url) or
-                    episode_filename == 'stream'):
+            if (youtube.is_video_link(self.url)
+                    or vimeo.is_video_link(self.url)
+                    or episode_filename == 'stream'):
                 episode_filename = self.title
 
             # If the basename is empty, use the md5 hexdigest of the URL
@@ -802,9 +818,7 @@ class PodcastEpisode(PodcastModelObject):
 
     @property
     def pubtime(self):
-        """
-        Returns published time as HHMM (or 0000 if not available)
-        """
+        """Return published time as HHMM (or 0000 if not available)."""
         try:
             return datetime.datetime.fromtimestamp(self.published).strftime('%H%M')
         except:
@@ -812,7 +826,7 @@ class PodcastEpisode(PodcastModelObject):
             return '0000'
 
     def playlist_title(self):
-        """Return a title for this episode in a playlist
+        """Return a title for this episode in a playlist.
 
         The title will be composed of the podcast name, the
         episode name and the publication date. The return
@@ -823,11 +837,18 @@ class PodcastEpisode(PodcastModelObject):
                 self.title,
                 self.cute_pubdate())
 
-    def cute_pubdate(self):
+    def cute_pubdate(self, show_time=False):
         result = util.format_date(self.published)
         if result is None:
             return '(%s)' % _('unknown')
-        else:
+
+        try:
+            if show_time:
+                timestamp = datetime.datetime.fromtimestamp(self.published)
+                return '<small>{}</small>\n{}'.format(timestamp.strftime('%H:%M'), result)
+            else:
+                return result
+        except:
             return result
 
     pubdate_prop = property(fget=cute_pubdate)
@@ -852,16 +873,17 @@ class PodcastEpisode(PodcastModelObject):
         return self.published_datetime().strftime('%y')
 
     def is_finished(self):
-        """Return True if this episode is considered "finished playing"
+        """Return True if this episode is considered "finished playing".
 
         An episode is considered "finished" when there is a
         current position mark on the track, and when the
         current position is greater than 99 percent of the
         total time or inside the last 10 seconds of a track.
         """
-        return (self.current_position > 0 and self.total_time > 0 and
-                (self.current_position + 10 >= self.total_time or
-                 self.current_position >= self.total_time * .99))
+        return (self.current_position > 0
+                and self.total_time > 0
+                and (self.current_position + 10 >= self.total_time
+                or self.current_position >= self.total_time * .99))
 
     def get_play_info_string(self, duration_only=False):
         duration = util.format_time(self.total_time)
@@ -909,11 +931,11 @@ class PodcastChannel(PodcastModelObject):
 
     feed_fetcher = gPodderFetcher()
 
-    def __init__(self, model, id=None):
+    def __init__(self, model, channel_id=None):
         self.parent = model
         self.children = []
 
-        self.id = id
+        self.id = channel_id
         self.url = None
         self.title = ''
         self.link = ''
@@ -978,7 +1000,7 @@ class PodcastChannel(PodcastModelObject):
         return new_url
 
     def check_download_folder(self):
-        """Check the download folder for externally-downloaded files
+        """Check the download folder for externally-downloaded files.
 
         This will try to assign downloaded files with episodes in the
         database.
@@ -1003,15 +1025,18 @@ class PodcastChannel(PodcastModelObject):
 
                 known_files.add(filename)
 
-        existing_files = set(filename for filename in
-                glob.glob(os.path.join(self.save_dir, '*'))
-                if not filename.endswith('.partial'))
+        # youtube-dl and yt-dlp create <name>.partial and <name>.partial.<ext> files while downloading.
+        # On startup, the latter is reported as an unknown external file.
+        # Both files are properly removed when the download completes.
+        existing_files = {filename
+                for filename in glob.glob(os.path.join(self.save_dir, '*'))
+                if not filename.endswith('.partial')}
 
         ignore_files = ['folder' + ext for ext in
                 coverart.CoverDownloader.EXTENSIONS]
 
-        external_files = existing_files.difference(list(known_files) +
-                [os.path.join(self.save_dir, ignore_file)
+        external_files = existing_files.difference(list(known_files)
+                + [os.path.join(self.save_dir, ignore_file)
                  for ignore_file in ignore_files])
         if not external_files:
             return
@@ -1085,7 +1110,7 @@ class PodcastChannel(PodcastModelObject):
 
             try:
                 tmp.update(max_episodes)
-            except Exception as e:
+            except Exception:
                 logger.debug('Fetch failed. Removing buggy feed.')
                 tmp.remove_downloaded()
                 tmp.delete()
@@ -1110,7 +1135,8 @@ class PodcastChannel(PodcastModelObject):
             return tmp
 
     def episode_factory(self, d):
-        """
+        """Create a PodcastEpisode from a dict.
+
         This function takes a dictionary containing key-value pairs for
         episodes and returns a new PodcastEpisode object that is connected
         to this object.
@@ -1207,7 +1233,7 @@ class PodcastChannel(PodcastModelObject):
                 next_feed = None
 
         # mark episodes not new
-        real_new_episode_count = 0
+        real_new_episodes = []
         # Search all entries for new episodes
         for episode in new_episodes:
             # Workaround for bug 340: If the episode has been
@@ -1219,17 +1245,18 @@ class PodcastChannel(PodcastModelObject):
                 episode.save()
 
             if episode.is_new:
-                real_new_episode_count += 1
+                real_new_episodes.append(episode)
 
             # Only allow a certain number of new episodes per update
-            if (self.download_strategy == PodcastChannel.STRATEGY_LATEST and
-                    real_new_episode_count > 1):
+            if (self.download_strategy == PodcastChannel.STRATEGY_LATEST
+                    and len(real_new_episodes) > 1):
                 episode.is_new = False
                 episode.save()
 
         self.children.extend(new_episodes)
 
         self.remove_unreachable_episodes(existing, seen_guids, max_episodes)
+        return real_new_episodes
 
     def remove_unreachable_episodes(self, existing, seen_guids, max_episodes):
         # Remove "unreachable" episodes - episodes that have not been
@@ -1237,8 +1264,8 @@ class PodcastChannel(PodcastModelObject):
         # Keep episodes that are currently being downloaded, though (bug 1534)
         if self.id is not None:
             episodes_to_purge = [e for e in existing if
-                    e.state != gpodder.STATE_DOWNLOADED and
-                    e.guid not in seen_guids and not e.downloading]
+                    e.state != gpodder.STATE_DOWNLOADED
+                    and e.guid not in seen_guids and not e.downloading]
 
             for episode in episodes_to_purge:
                 logger.debug('Episode removed from feed: %s (%s)',
@@ -1251,8 +1278,8 @@ class PodcastChannel(PodcastModelObject):
                     self.children.remove(episode)
 
         # This *might* cause episodes to be skipped if there were more than
-        # max_episodes_per_feed items added to the feed between updates.
-        # The benefit is that it prevents old episodes from apearing as new
+        # limit.episodes items added to the feed between updates.
+        # The benefit is that it prevents old episodes from appearing as new
         # in certain situations (see bug #340).
         self.db.purge(max_episodes, self.id)  # TODO: Remove from self.children!
 
@@ -1261,21 +1288,22 @@ class PodcastChannel(PodcastModelObject):
 
     def update(self, max_episodes=0):
         max_episodes = int(max_episodes)
+        new_episodes = []
         try:
             result = self.feed_fetcher.fetch_channel(self, max_episodes)
 
             if result.status == feedcore.UPDATED_FEED:
-                self._consume_updated_feed(result.feed, max_episodes)
+                new_episodes = self._consume_updated_feed(result.feed, max_episodes)
             elif result.status == feedcore.NEW_LOCATION:
                 # FIXME: could return the feed because in autodiscovery it is parsed already
                 url = result.feed
                 logger.info('New feed location: %s => %s', self.url, url)
-                if url in set(x.url for x in self.model.get_podcasts()):
+                if url in {x.url for x in self.model.get_podcasts()}:
                     raise Exception('Already subscribed to ' + url)
                 self.url = url
                 # With the updated URL, fetch the feed again
                 self.update(max_episodes)
-                return
+                return new_episodes
             elif result.status == feedcore.NOT_MODIFIED:
                 pass
 
@@ -1302,6 +1330,7 @@ class PodcastChannel(PodcastModelObject):
         self._determine_common_prefix()
 
         self.db.commit()
+        return new_episodes
 
     def delete(self):
         self.db.delete_podcast(self)
@@ -1409,8 +1438,8 @@ class PodcastChannel(PodcastModelObject):
         download_folder = download_folder.strip('.' + string.whitespace)
 
         for folder_name in util.generate_names(download_folder):
-            if (not self.db.podcast_download_folder_exists(folder_name) or
-                    self.download_folder == folder_name):
+            if (not self.db.podcast_download_folder_exists(folder_name)
+                    or self.download_folder == folder_name):
                 return folder_name
 
     def get_save_dir(self, force_new=False):
@@ -1511,9 +1540,9 @@ class Model(object):
 
     @classmethod
     def sort_episodes_by_pubdate(cls, episodes, reverse=False):
-        """Sort a list of PodcastEpisode objects chronologically
+        """Sort a list of PodcastEpisode objects chronologically.
 
-        Returns a iterable, sorted sequence of the episodes
+        Returns a iterable, sorted sequence of the episodes.
         """
         return sorted(episodes, key=cls.episode_sort_key, reverse=reverse)
 
