@@ -106,7 +106,8 @@ def get_track_length(filename):
 
 
 def episode_filename_on_device(config, episode):
-    """
+    """Return the basename of the episode file to save on device.
+
     :param gpodder.config.Config config: configuration (for sync options)
     :param gpodder.model.PodcastEpisode episode: episode to get filename for
     :return str: basename minus extension to use to save episode on device
@@ -130,7 +131,8 @@ def episode_filename_on_device(config, episode):
 
 
 def episode_foldername_on_device(config, episode):
-    """
+    """Return the folder name to which the episode is saved on device.
+
     :param gpodder.config.Config config: configuration (for sync options)
     :param gpodder.model.PodcastEpisode episode: episode to get folder name for
     :return str: folder name to save episode to on device
@@ -146,9 +148,9 @@ def episode_foldername_on_device(config, episode):
 
 
 class SyncTrack(object):
-    """
-    This represents a track that is on a device. You need
-    to specify at least the following keyword arguments,
+    """Class representing a track that is on a device.
+
+    You need to specify at least the following keyword arguments,
     because these will be used to display the track in the
     GUI. All other keyword arguments are optional and can
     be used to reference internal objects, etc... See the
@@ -162,6 +164,7 @@ class SyncTrack(object):
     passed to the function (the values will default to None
     for all required fields).
     """
+
     def __init__(self, title, length, modified, **kwargs):
         self.title = title
         self.length = length
@@ -528,7 +531,11 @@ class MP3PlayerDevice(Device):
         to_file_exists = to_file.query_exists()
         from_size = episode.file_size
         to_size = episode.file_size
-        if to_file_exists:
+        # An interrupted sync results in a partial file on the device that must be removed to fully sync it.
+        # Comparing file size would detect such files and finish uploading.
+        # However, some devices add metadata to files, increasing their size, and forcing an upload on every sync.
+        # File size and checksum can not be used.
+        if to_file_exists and self._config.device_sync.compare_episode_filesize:
             try:
                 info = to_file.query_info(Gio.FILE_ATTRIBUTE_STANDARD_SIZE, Gio.FileQueryInfoFlags.NONE)
                 to_size = info.get_attribute_uint64(Gio.FILE_ATTRIBUTE_STANDARD_SIZE)
@@ -578,10 +585,13 @@ class MP3PlayerDevice(Device):
             if self._config.one_folder_per_podcast:
                 if path_info.get_file_type() == Gio.FileType.DIRECTORY:
                     path_file = root_path.get_child(path_info.get_name())
-                    for child_info in path_file.enumerate_children(attributes, Gio.FileQueryInfoFlags.NONE, None):
-                        if child_info.get_file_type() == Gio.FileType.REGULAR:
-                            child_file = path_file.get_child(child_info.get_name())
-                            self.add_sync_track(tracks, child_file, child_info, path_info.get_name())
+                    try:
+                        for child_info in path_file.enumerate_children(attributes, Gio.FileQueryInfoFlags.NONE, None):
+                            if child_info.get_file_type() == Gio.FileType.REGULAR:
+                                child_file = path_file.get_child(child_info.get_name())
+                                self.add_sync_track(tracks, child_file, child_info, path_info.get_name())
+                    except GLib.Error as err:
+                        logger.error('get all tracks for %s failed: %s', path_file.get_uri(), err.message)
 
             else:
                 if path_info.get_file_type() == Gio.FileTypeFlags.REGULAR:
@@ -630,10 +640,12 @@ class MP3PlayerDevice(Device):
         return True
 
 
-class SyncCancelledException(Exception): pass
+class SyncCancelledException(Exception):
+    pass
 
 
-class SyncFailedException(Exception): pass
+class SyncFailedException(Exception):
+    pass
 
 
 class SyncTask(download.DownloadTask):
@@ -762,7 +774,7 @@ class SyncTask(download.DownloadTask):
     def __enter__(self):
         return self.__lock.acquire()
 
-    def __exit__(self, type, value, traceback):
+    def __exit__(self, exception_type, value, traceback):
         self.__lock.release()
 
     def notify_as_finished(self):
@@ -836,7 +848,7 @@ class SyncTask(download.DownloadTask):
         try:
             logger.info('Starting SyncTask')
             self.device.add_track(self, reporthook=self.status_updated)
-        except SyncCancelledException as e:
+        except SyncCancelledException:
             sync_result = SyncTask.CANCELLED
         except Exception as e:
             sync_result = SyncTask.FAILED
